/*

   $Id: libcraw++.cpp,v 1.2 2005/03/24 05:41:04 dmg Exp $

   Library to read the records from a RAW file from Canon

   Credits:

     David Coffin:  Version 0.1 of this program is based on his parse.c
     David Burren:  Embedded exif record 
     Peter Galbavy: His documentation of the format helped a lot


 * Copyright (C) 2004 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "libcraw.h"
#include "libciff.h"

/* Heap file header */
typedef struct  {
   int2     byteOrder; /* MM Motorola ---big endian, II Intel --little endian*/
   int4     length;
   char     identifier[8]; /* type HEAP, subtype heap  CCDR */
   int2     version;
   int2     subversion;
   char     unused[14]; /* reserved 8 */
} raw_header;



/*
 * Internal record
 */

typedef struct {
   int   typeRecord;
   int   length;
   int   recordOffset;
   int   fileOffset;
} raw_record_head_struct;

typedef struct {
   int type;
   int length;
   long recordOffset;
} photo_record;

typedef struct {
   int2   type;
   int4   length;
   int4   recordOffset;
} raw_record;
   

/* GLobal variables */

int debug=0;


int Find_Photo_Records_Offset(craw_photo_type *photo)
{
   int  offset;
   fseek(photo->file, -4, SEEK_END);

   offset = Read_Int_4(photo->file, photo->byteOrder);
   offset += photo->headerLength;

   if (debug) {
      fprintf(ERROR_OUTPUT, "Offset for records: %d\n", offset);
   }
   return offset;
}

craw_photo_type *craw_Read_Header(FILE *photoFile)
{
   raw_header rawHeader;
   craw_photo_type *pHeader;
   if ((pHeader = (craw_photo_type *)calloc(sizeof(craw_photo_type), 1)) == NULL) {
      My_Die("Not enough memory\n");
   }
   
   /* Seek to the beginning of the file to make sure */
   
   fseek(photoFile, 0, SEEK_SET);
   
   assert(sizeof(raw_header) == HEADER_LENGTH);
   
   if (fread (&rawHeader, 1, HEADER_LENGTH, photoFile) != HEADER_LENGTH) {
      My_Die("Read incomplete header!");
   }
   /* 
      Move data to the photo header struct, that takes care of
      endianism
   */
   if (rawHeader.byteOrder.value == LITTLE_ENDIAN_ENCODING) {
      pHeader->byteOrder = little_endian;
   } else {
      pHeader->byteOrder = big_endian;
   }

   pHeader->headerLength = Convert_Number_4_bytes(rawHeader.length, pHeader->byteOrder);
   
   /* Copy identifier and rest of the data */
   memcpy(&(pHeader->identifier), rawHeader.identifier, HEADER_IDENTIFIER_LENGTH);
   pHeader->identifier[8] = '\0';

   pHeader->file = photoFile;
   
   fseek(photoFile, 0, SEEK_END);
   
   pHeader->fileLength =  ftell(photoFile);

   return pHeader;
}

craw_photo_type *craw_Load_Photo(FILE *photoFile) 
{
   craw_photo_type *photo;

   photo = craw_Read_Header(photoFile);
   if (photo == NULL )
      goto justreturn;
   

   if (!memcmp(HEAP_CCDR_IDENT, photo->identifier, 8) == 0 &&
       !memcmp(HEAP_CR2_IDENT, photo->identifier, 8) == 0 ) {
      fprintf(stderr, "This file is not a Canon raw photo\n");
      for (int i=0;i<8;i++) {
	fprintf(stderr, "Byte %d [%c] [%x] [%x] [%c]\n", i, photo->identifier[i], photo->identifier[i], (unsigned char)HEAP_CR2_IDENT[i],(unsigned char)HEAP_CR2_IDENT[i]);
      }
      goto freephoto;
   }
   
#ifdef __cplusplus
   Ciff_Parser* ciffParser;
   ciffParser = new Ciff_Parser();
   ciffParser->Set_Photo(photo);
   ciffParser->Parse();
#else
   ciff_Read_Data(photo);  asdfasfd
#endif

   return photo;

   /* Error handlers start here */


 freephoto:
   free(photo);
   photo = NULL;
 justreturn:
   return photo;
   
}
   


#define RECORD_RAW               0x2005
#define RECORD_JPEG              0x2007

/* dataOffset is the data offset of the heap
   lengthData is the length of the heap */

int craw_Find_Jpeg_Location_In_Heap(craw_photo_type *pPhoto, 
				    long dataOffset, 
				    long lengthData)
{
   int count = 0;
   int i;
   long recordsOffset;

   assert(dataOffset > 0);
   assert(lengthData > 0);

   /* read the last 4 bytes from the data block */
   fseek(pPhoto->file, dataOffset+lengthData - 4, SEEK_SET);

   recordsOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

   /*
   if (debug) fprintf(ERROR_OUTPUT, "%sOffset of record header %ld %lx\n", prefix, recordsOffset,recordsOffset);
   */

   /*   photo_record *pRecords; */
   
   /* First 2 bytes are the number of records */

   fseek(pPhoto->file, recordsOffset + dataOffset, SEEK_SET);

   /*
   if (debug) fprintf(ERROR_OUTPUT, "%sSeek at %ld %lx\n", prefix, recordsOffset+dataOffset,recordsOffset+dataOffset);
   */

   count = Read_Int_2(pPhoto->file, pPhoto->byteOrder);

   if (debug) {
      fprintf(ERROR_OUTPUT, ">Records [%d] %x\n", count, count);
   }
   /* let us make sure we don't have weird values */
   assert(count> 0 && count <100);

   /* Allocate the records */

   for (i=0;i<count;i++) {
      int type =-1;
      long int length=-1;
      long int recordOffset=-1;
      

      type = Read_Int_2(pPhoto->file, pPhoto->byteOrder);
      if (type & 0x4000) {
	 fprintf(ERROR_OUTPUT, "This should not be reached\n");
	 assert(0);
      } else {
	 length = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
      }



      switch (type) {
      case RECORD_JPEG:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 
	 pPhoto->jpegOffset = recordOffset;
	 pPhoto->jpegLength = length;
	 
	 fprintf(ERROR_OUTPUT, "Embedded jpeg image %ld %ld\n", recordOffset, length);
	 return 1;
      case RECORD_RAW:
	 /* We need to skip the next 4 bytes */
	 Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	 
	 break;
      default:
	 fprintf(ERROR_OUTPUT, "REading record %5d Type [%5x] type [%5d] length %8ld %8lx\n",
		 i, type, type, length, length);
	 
	 fprintf(ERROR_OUTPUT, "Illegal Canon Raw file\n");
	 return 0;
      }
   }
   fprintf(ERROR_OUTPUT, "Embedded jpeg image not found\n");
   return 0;
}

void craw_Destroy_Photo(craw_photo_type *photo)
{
   /* I think I don't allocate anything but the memory for the 
      record */
   free(photo);
}



int craw_Get_Jpeg_Location(craw_photo_type *photo, 
			   unsigned long *pDataOffset, 
			   unsigned long *pDataLength)
{
   if (photo->jpegOffset > 0 && photo->jpegLength) {
      *pDataOffset = photo->jpegOffset;
      *pDataLength = photo->jpegLength;
      return 1;
   } else
      return 0;
}
