/*

   $Id: libraw.c,v 1.7 2005/03/24 06:12:15 dmg Exp $

   Library to deal with endianism

 * Copyright (C) 2004 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libraw.h"

void My_Die(char *message)
{
   fprintf(ERROR_OUTPUT, message);
   exit(1);
}

int Convert_Number_4_bytes(int4 value, byte_order endianism)
{
   if (endianism == little_endian || endianism == unchanged) {
      return value.bytes[0] + (value.bytes[1] << 8) + (value.bytes[2] << 16) + (value.bytes[3] << 24);
   } else {
      return (value.bytes[0] << 24) + (value.bytes[1] << 16) + (value.bytes[2] << 8) + value.bytes[0];
   }
}

float Convert_Number_Float4(float4 value, byte_order endianism)
{
/* Floats are totally inverted in each architecture 

The following are the values (int) 0x12345678 and (float)1.23123e+07
written in each architecture

[dmg@selenium:/tmp] hexdump motorola
0000000 1234 5678 4b3b df01
0000008
[dmg@selenium:/tmp] hexdump intel
0000000 7856 3412 01df 3b4b
0000008

*/
   float *pValue = (float*)&value;
   float test = 1.0;
   float4 *pTest = (float4*)&test;
   
   if (pTest->bytes[3] == 0 && pTest->bytes[2] == 0) {

      /* We are motorola. */
#define MY_SWAP(a, b) (temp = a, a = b, b = temp)

      if (endianism == little_endian || endianism == unchanged) {
	 /* Intel */
	 /* Swap bytes */
	 char temp;
	 MY_SWAP(value.bytes[0],value.bytes[3]);
	 MY_SWAP(value.bytes[1],value.bytes[2]);
      } else {
	 /* Do nothing */
      }
				  
   } else {
      /* We are intel */
      if (endianism == little_endian || endianism == unchanged) {
	 /* Intel */
	 /* Do nothing */
      } else {
	 /* Swap bytes */
	 char temp;
	 MY_SWAP(value.bytes[0],value.bytes[3]);
	 MY_SWAP(value.bytes[1],value.bytes[2]);
      }
   }

#undef MY_SWAP
   return *pValue;
}

byte_order Host_Endianism(void)
{
   int i = 1;
   char *p = (char*)&i;
   if (*p == 1) 
      return little_endian;
   else
      return big_endian;

}


int Convert_Number_2_bytes(int2 value, byte_order endianism)
{
   if (endianism == little_endian || endianism == unchanged) {
      return value.bytes[0] + (value.bytes[1] << 8);
   } else {
      return (value.bytes[0] << 8) + value.bytes[1];
   }
}

int Read_Int_2(FILE *file, byte_order endianism) 
{
   int2 value;

   if (fread (&value, 1, 2, file) != 2) {
      My_Die("Read error. I could not read 2 bytes");
   }
   
   return Convert_Number_2_bytes(value, endianism);
}


int Read_Int_4(FILE *file, byte_order endianism) 
{
   int4 value;

   if (fread (&value, 1, 4, file) != 4) {
      My_Die("Read error. I could not read 4 bytes");
   }
   
   return Convert_Number_4_bytes(value, endianism);
}


void Print_Int4(void *buffer, int length, byte_order endianism)
{
   int i;
   unsigned long lValue;
   int4 value;
   for (i=0;i<length;i+=4) {
      memcpy(&value, (char*)buffer + i, 4);
      lValue = Convert_Number_4_bytes(value, endianism);
      fprintf(stderr,"                            %d %d int4 values %lu %lx\n",i, i/4, lValue, lValue);
   }
}

void Print_Int2(void *buffer, int length, byte_order endianism)
{
   int i;
   long lValue;
   int2 value;
   for (i=0;i<length;i+=2) {
      memcpy(&value, (char*)buffer +i, 2);
      lValue = Convert_Number_2_bytes(value, endianism);
      fprintf(stderr,"                            %d i %d int2 values %ld %lx %d\n",i, i/2,lValue, lValue, (signed short)lValue);
   }
}


void Print_Bytes(unsigned char *buffer, int length)
{
   int i;
   for (i=0;i<length;i++) {
      fprintf(stderr,"                            %d %x values %d %x\n",i, i, buffer[i], buffer[i]);
   }
}

int Fix_Endianism_In_Record(void *record, char* recordSpec, int length, byte_order byteOrder)
{
  /* Record spec is a NULL terminated string that contains for each
     field a specification of how it must be translated

     4 -> int4
     2 -> int2
     c -> char (unchanged)
     f -> float4
     We fix in-situ
  */


  /* I am not sure that the an increment of 1 to a void is equal to 1 byte, so let us assert */
  assert((int)record + 1 == (int)(record + 1));

  char *currentField = recordSpec;
  int currentLength = 0;

  /* ANd just for the sake of safety, let us make sure data types are the size
     we need them */

  assert(sizeof(int) == 4);
  assert(sizeof(short) == 2);
  assert(sizeof(float) == 4);
  assert(sizeof(char) == 1);


  //  printf("Record %s length %d\n", recordSpec, length);
  while (*currentField != '\0') {

    switch (*currentField) {
    case '4':
      /* Make sure we have space */
      if (currentLength +4 <= length) {
	*(int*)record = Convert_Number_4_bytes(*(int4*)record, byteOrder);
	record += 4;
	currentLength+=4;
      }
      break;
    case 'f':
      if (currentLength +4 <= length) {
	*(float*)record = Convert_Number_Float4(*(float4*)record, byteOrder);
	record += 4;
	currentLength+=4;
      }
      break;
    case '2':
      /* Make sure we have space */
      //      printf("length %d, current length %d %c\n", length, currentLength, *currentField);
      if (currentLength +2 <= length) {
	*(short*)record = Convert_Number_2_bytes(*(int2*)record, byteOrder);
	record += 2;
	currentLength+=2;
      }
      break;
    case 'c':
      if (currentLength +1 <= length) {
	record += 1;
	currentLength+=1;
      }
      break;
    default:
      fprintf(stderr, "Illegal  specification in record spec [%s] char [%c]\n", recordSpec,
	      *currentField);
      exit(1);
    }
    currentField++;
  }

  return currentLength;

}

